/*
    Copyright 2008-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 3 of the License.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file gpl-3.0.txt. If not, see
    <http://www.gnu.org/licenses/>.
*/
#ifndef EMEA_H
#define EMEA_H

#include <QtGui/QMainWindow>
#include <QtCore/QModelIndex>
#include <QtXml/QDomDocument>

#include <thread_template.h>
#include <spectrumgui.h>

namespace NSCore {
	class DataSetProducer;
	class DataSetChainMember;
};

/**
 * @brief Generic framework for various evaluation apps.
 *
 * EMEA stands for Epos Modularized Evaluation Applications.
 *
 * @todo docs
 */
class EMEAWindow : public QMainWindow
{
	Q_OBJECT
	public:
		EMEAWindow( QWidget* );
		~EMEAWindow();

		void closeEvent( QCloseEvent* );

	protected:
		/**
		 * @brief initialize the chain of evaluation
		 *
		 * This function is called at the right time during startup to
		 * initialize the needed chain of producers and chain-members for the
		 * given task.
		 *
		 * You compiler will tell you that you have to implement this function.
		 * Trust him.
		 *
		 * Use addToChainStart() and addToChain() to add elements to the chain
		 * and use addWidgetToStack() to add widgets to the stacked gui.
		 */
		virtual void setupChain( int threads ) =0;

		/**
		 * @brief Start the evaluation.
		 *
		 * The app was told ( via commandline ) to start the evaluation by itself. Do this here...
		 */
		virtual void autoStart() {}

		/**
		 * @brief Add elements to the start of the chain.
		 */
		void addToChainStart( NSCore::DataSetProducer* );
		/**
		 * @{
		 *
		 * @brief Add elements to the chain
		 *
		 * Each call will create a new level in the chain. The elements of each
		 * level are connected with all elements of the previous (thus
		 * producing) level and with all elements of the next (consuming) level.
		 */
		void addToChain( NSCore::DataSetChainMember* );
		void addToChain( QList<NSCore::DataSetChainMember*> );
		// @}

		/**
		 * @brief Add a signal to the logging
		 */
		void addLogging( QString name, QObject*, const char* );

		/**
		 * @brief Add the threads to a list that gets stopped automaticly
		 */
		void addThread( QThread* );
		/**
		 * @brief instantiate an object in its own thread and return the pointer
		 *
		 * This one should make multi-thread emea apps most easy. Simply create
		 * the object via this template function and you get threads for free.
		 *
		 * You have to add the objects to the chain by hand.
		 */
		template<class ThreadClass>
		ThreadClass* objectInNewThread() {
			NSCore::Thread_Template<ThreadClass>* t = new NSCore::Thread_Template<ThreadClass>( this );
			t->start();
			t->waitForStarted();
			addThread( t );
			return t->item();
		}

		/**
		 * @brief the bridge to create Spectrum-Guis for new Spectra.
		 */
		SpectrumGuiCreator* spectrumGuiCreator();

		/**
		 * @brief The arguments used to start the app
		 */
		QStringList arguments();

	protected slots:
		/**
		 * @brief Add a widget to the stacked gui
		 *
		 * When the widget has an object name (set with setObjectName()) this
		 * will be used, if it has an icon (set with setWindowIcon()) this will
		 * be used too.
		 *
		 * @param w The widget to add.
		 * @param scrollarea Whether the widget should be added inside a QScrollArea. Defaults to false.
		 */
		void addWidgetToStack( QWidget* w, bool scrollarea = false );

	private slots:
		/**
		 * Do the real setup after all the constructors are run. C++ forbids to
		 * call the virtual setupChain() from the constructor of EMEAWindow, so
		 * we schedule lateInit() first after entering the event-loop.
		 */
		void lateInit();

		/**
		 * Execute the autoStart() with some time-delay so things can settle
		 */
		void delayedAutoStart();

		/**
		 * @brief Switch the connection to the osci
		 */
		void toggleOsci( bool );

		/**
		 * @brief Switch the connection to the result-viewer
		 */
		void toggleViewer( bool );

		void loadXmlFromFile();
		void loadXmlFromFile( QString );
		void loadXml( QDomDocument );

		void saveXmlToFile();
		void saveXml( QDomNode, QDomDocument );

	private:
		class Private;
		Private* d;
};

#endif // EMEA_H
