/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DATASETCHAIN_H
#define DATASETCHAIN_H

#include <list>

#include <datatypes.h>
#include <pointer_template.h>

#include <qtreeitem.h>

class QTimer;
class QReadWriteLock;

namespace NSCore {

	class DataSetProducer;
	class DataSetChainMember;

	/**
	 * @brief Baseclass for Producers
	 *
	 * Subclass to produce AcquisitionDataSets and feed them into a chain.
	 */
	class DataSetProducer : public QTreeItem
	{
		Q_OBJECT
		public:
			/**
			 * @brief empty constructor
			 */
			DataSetProducer( QObject* =0 );
			/**
			 * @brief destructor
			 *
			 * Deregisters itself at the consumers.
			 */
			~DataSetProducer();

			/**
			 * @brief Register a consumer
			 *
			 * This registers a DataSetChainMember as this objects consumer.
			 *
			 * The emitted data is sent to all consumers via the (internal)
			 * DataSetEvent through Qt's event system.
			 */
			void registerConsumer( DataSetChainMember* );
			/**
			 * @brief Unregister a consumer
			 */
			void unregisterConsumer( DataSetChainMember* );
			/**
			 * @brief Unregister all consumers
			 */
			void unregisterAllConsumers();

			enum Dispatching { DispatchToAll, DispatchEqual };
			/**
			 */
			void setDispatching( Dispatching );

			/**
			 * @brief Return the number of registered consumers.
			 *
			 * This way a producer can decide not to do expensive conversions
			 * and similar stuff when there is no consumer to actually use the
			 * data.
			 */
			int consumersSize() const;
			/**
			 * @brief Return wether there are consumers or not.
			 *
			 * Even simplier then consumersSize().
			 */
			bool consumersIsEmpty() const;

		signals:

			/**
			 * @brief Speed-o-meter
			 *
			 * The enqueue()-function tries to keep track on how many datasets
			 * where sent to the next chain-member and will emit this signal at
			 * (ir-)regular intervals.
			 */
			void countsPerSecond( double );

		protected:
			/**
			 * @brief Add data to the queue
			 *
			 * The data is sent to all registered consumers via Qt' event
			 * system.
			 */
			void enqueue( AcqisitionDataSet& );

		private:
			/**
			 * Private class for the internals
			 * @internal
			 */
			class DataSetProducerPrivate;
			/**
			 * Internal pointer
			 * @internal
			 */
			DataSetProducerPrivate* _d;
	};

	/**
	 * @brief Consume datasets, do something with them and give them to the next consumer
	 *
	 * Formerly there was producer and consumer and the chainmember to combine
	 * both. But consumers alone are stupid, even if things like the osci don't
	 * modify the data, they aren't always the last in the chain.
	 *
	 * So the DataSetChainMember is the base class for elements consuming
	 * datasets, doing something with them and reaching them further down the
	 * chain.
	 */
	class DataSetChainMember : public DataSetProducer
	{
		Q_OBJECT

		public:
			/// Defaultconstructor
			DataSetChainMember( QObject* =0 );
			/// Deconstructor
			~DataSetChainMember();

			/**
			 * @brief intercept to get events
			 *
			 * By this function we get notified of DataSetEvent signaling new
			 * data...
			 */
			bool event( QEvent* );

		protected:
			/**
			 * @brief Data arrived!
			 *
			 * Subclass this function to handle the data. Return true when the
			 * data should be sent to the next in the chain. Otherwise it gets
			 * deleted.
			 */
			virtual bool dataArrived( NSCore::AcqisitionDataSet& ) =0;
	};

};

#endif // DATASETCHAIN_H

