/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "datatypes.h"

#include "namedparametertreeitem.h"

#include <cmath>
#include <QtCore/QMutex>
#include <QtCore/QMutexLocker>
#include <QtCore/QSemaphore>

namespace NSCore {

	class MemoryPool
	{
		private:
			MemoryPool() {
				activeDataSets.release( 1000 );
			}
			MemoryPool( MemoryPool& ) {}
			MemoryPool( MemoryPool* ) {}
			~MemoryPool() {
				/*
				 * Will this ever get executed???
				 */
				while ( ! chunkPool.empty() ) {
					int size = chunkPool.begin()->first;
					qDebug() << " Deleting all elements of size" << size;
					std::multimap<int,double*>::const_iterator begin = chunkPool.lower_bound( size );
					std::multimap<int,double*>::const_iterator end = chunkPool.upper_bound( size );
					while ( begin != end ) {
						delete[] begin->second;
						begin++;
					}
					chunkPool.erase( size );
				}
			}

			QMutex lock;
			typedef std::multimap<int, double*> doublemap;
			typedef std::pair<int, double*> doublepair;
			doublemap chunkPool;
			doublemap minmaxPool;

		public:
			static MemoryPool* the() {
				static MemoryPool* me = new MemoryPool();
				return me;
			}

			QSemaphore activeDataSets;

			double* getMem( int size ) {
				//qDebug() << "MemoryPool::getMem(" << size << ")";
				QMutexLocker l( &lock );
				double* tmp = 0;
				if ( chunkPool.count( size ) > 0 ) {
					doublemap::iterator it = chunkPool.find( size );
					tmp = it->second;
					chunkPool.erase( it );
					//qDebug() << "MemoryPool::getMem() getting a chunk out of the pool";
				} else {
					//qDebug() << "MemoryPool::getMem(" << size << ") have to allocate new space\n  chunkPool.size() is" << chunkPool.size() << "\n  activeChunks.available() is" << activeChunks.available();
					tmp = new double[ size ];
				}
				return tmp;
			}
			void freeMem( double* mem, int size ) {
				QMutexLocker l( &lock );
				//qDebug() << "MemoryPool::freeMem(" << mem << ")";
				if ( chunkPool.size() > 10000 )
					delete[] mem;
				else
					chunkPool.insert( doublepair( size, mem ) );
			}
			double* getMinMax( int size ) {
				QMutexLocker l( &lock );
				double* tmp = 0;
				if (minmaxPool.count( size ) > 0 ) {
					doublemap::iterator it = minmaxPool.find( size );
					tmp = it->second;
					minmaxPool.erase( it );
				} else {
					tmp = new double[ size ];
				}
				return tmp;
			}
			void freeMinMax( double* mem, int size ) {
				QMutexLocker l( &lock );
				if ( minmaxPool.size() > 10000 )
					delete[] mem;
				else
					minmaxPool.insert( doublepair( size, mem ) );
			}
	};



	struct AcqisitionDataSet::AcqisitionDataSet_Private
	{
#if QT_VERSION >= 0x040400
		QAtomicInt ref;
#else
		QAtomic ref;
#endif

		NamedQVariantTreeItem* resulttree;

		struct RealDataSet {
			// The data itself
			int channels;
			int samples;
			double* data;

			// caches
			double* min;
			double* max;

			// generations
			int generationNumber;
			QString generationPrefix;

			/// constructor
			RealDataSet( const int& _channels, const int& _samples, const int& _generationNumber, const QString& _generationPrefix )
				: channels( _channels )
				, samples( _samples )
				, generationNumber( _generationNumber )
				, generationPrefix( _generationPrefix )
			{
				data = MemoryPool::the()->getMem( channels*samples );
				min = new double[ channels ];
				max = new double[ channels ];
				for ( int i=0; i<channels; ++i ) {
					min[i] = INFINITY;
					max[i] = -INFINITY;
				}
			}
			~RealDataSet() {
				MemoryPool::the()->freeMem( data, channels * samples );
				delete[] min;
				delete[] max;
			}
			private:
			RealDataSet( const RealDataSet &copy ) {
				qWarning() << "Trying to copy a RealDataSet!!!";
			}
		};
		QList<RealDataSet*> datasets;

		struct cache_t {
			int currentGeneration;
			QString currentPrefix;
		} cache;

		AcqisitionDataSet_Private()
			: resulttree( new NamedQVariantTreeItem( "", QVariant::Invalid, 0 ) )
		{
			MemoryPool::the()->activeDataSets.acquire();
			/*if ( ! MemoryPool::the()->activeDataSets.tryAcquire() ) {
				qWarning() << "AcqisitionDataSet_Private: Have to wait for a semaphore!" << endl;
				MemoryPool::the()->activeDataSets.acquire();
			}*/
		}
		~AcqisitionDataSet_Private() {
			delete resulttree;
			/*foreach( RealDataSet tmp, datasets ) {
				MemoryPool::the()->freeMem( tmp.data, tmp.channels * tmp.samples );
				tmp.data = 0;
				MemoryPool::the()->freeMinMax( tmp.min, tmp.channels );
				MemoryPool::the()->freeMinMax( tmp.max, tmp.channels );
			}*/
			while ( ! datasets.isEmpty() )
				delete datasets.takeFirst();
			MemoryPool::the()->activeDataSets.release();
		}
		private:
		AcqisitionDataSet_Private( const AcqisitionDataSet_Private& ) {}
	};

	AcqisitionDataSet::AcqisitionDataSet( int channels, int samples )
		: d( new AcqisitionDataSet_Private )
	{
		//qDebug() << "AcqisitionDataSet::AcqisitionDataSet(" << channels << "," << samples << ")";
		d->ref.ref();
		newGeneration( "data", channels, samples );
	}
	AcqisitionDataSet::AcqisitionDataSet( const AcqisitionDataSet& c )
		: d( c.d )
	{
		//qDebug() << "AcqisitionDataSet::AcqisitionDataSet(" << &c << ") copy-constructor";
		d->ref.ref();
	}
	AcqisitionDataSet::~AcqisitionDataSet() {
		//qDebug() << "AcqisitionDataSet::~AcqisitionDataSet()";
		if ( ! d->ref.deref() )
			delete d;
	}

	int AcqisitionDataSet::currentGenerationNumber() const {
		return d->cache.currentGeneration;
	}
	void AcqisitionDataSet::debugPrint() const {
		d->resulttree->debugPrint();
	}

	void AcqisitionDataSet::newGeneration( QString prefix, int channels, int samples ) {
		//qDebug() << "AcqisitionDataSet::newGeneration(" << prefix << "," << channels << "," << samples << ")";
		pushPrefix( prefix );
		AcqisitionDataSet_Private::RealDataSet* tmp = new AcqisitionDataSet_Private::RealDataSet( channels, samples, d->datasets.size(), d->cache.currentPrefix );
		d->cache.currentGeneration = d->datasets.size();
		d->datasets.push_back( tmp );
	}

	void AcqisitionDataSet::pushPrefix( QString n ) {
		//qDebug() << "AcqisitionDataSet::pushPrefix(" << n << ")";
		d->cache.currentPrefix += "/" + n;
		d->resulttree->addQVariant( d->cache.currentPrefix, QVariant::Invalid );
	}
	void AcqisitionDataSet::popPrefix( QString n ) {
		//qDebug() << "AcqisitionDataSet::popPrefix(" << n << ")";
		if ( d->cache.currentPrefix.endsWith( n ) ) {
			d->cache.currentPrefix.chop( n.size() + 1 );
			int generation = -1;
			QString prefix = d->cache.currentPrefix;
			while ( generation == -1 && prefix.size() > 0 ) {
				foreach( AcqisitionDataSet_Private::RealDataSet* tmp, d->datasets ) {
					if ( tmp->generationPrefix == prefix )
						generation = tmp->generationNumber;
				}
				prefix = prefix.section( "/", 0, -2 );
			}
			d->cache.currentGeneration = generation;
		}
	}

	QString AcqisitionDataSet::getPrefix( int generation ) const {
		if ( generation == -1 || generation == d->cache.currentGeneration )
			return d->cache.currentPrefix;

		QString ret;
		for ( int i=0; ret=="" && i<d->datasets.size(); ++i ) {
			if ( d->datasets.at( i )->generationNumber == generation )
				ret = d->datasets.at( i )->generationPrefix;
		}
		return ret;
	}

	double AcqisitionDataSet::sample( int channel, int sample, int generation ) const {
		if ( generation == -1 )
			generation = d->cache.currentGeneration;
		AcqisitionDataSet_Private::RealDataSet* tmp =  d->datasets[ generation ];
		double ret = 0;
		if ( sample >=0 && sample < tmp->samples )
			ret = tmp->data[ channel*tmp->samples + sample ];
		//qDebug() << "AcqisitionDataSet::sample(" << channel << "," << sample << "," << generation << ") returning" << ret;
		return ret;
	}
	void AcqisitionDataSet::setSample( int channel, int sample, double value, int generation ) {
		//qDebug() << "AcqisitionDataSet::setSample(" << channel << "," << sample << "," << value << "," << generation << ")";
		if ( generation == -1 )
			generation = d->cache.currentGeneration;
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		tmp->data[ channel*tmp->samples + sample ] = value;
		if ( value > tmp->max[ channel ] ) {
			tmp->max[ channel ] = value;
			addResult( QString( "channel %1/max" ).arg( channel ), value );
			addResult( QString( "channel %1/max at" ).arg( channel ), sample );
		}
		if ( value < tmp->min[ channel ] ) {
			tmp->min[ channel ] = value;
			addResult( QString( "channel %1/min" ).arg( channel ), value );
			addResult( QString( "channel %1/min at" ).arg( channel ), sample );
		}
	}
	int AcqisitionDataSet::samples( int generation ) const {
		if ( generation == -1 )
			generation = d->cache.currentGeneration;
		return d->datasets[ generation ]->samples;
	}
	int AcqisitionDataSet::channels( int generation ) const {
		if ( generation == -1 )
			generation = d->cache.currentGeneration;
		return d->datasets[ generation ]->channels;
	}

	double* AcqisitionDataSet::getRawChannel( int channel, int generation ) {
		if ( generation == -1 )
			generation = d->cache.currentGeneration;
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		return &( tmp->data[ channel*tmp->samples ] );
	}
	double* AcqisitionDataSet::getRawData( int generation ) {
		if ( generation == -1 )
			generation = d->cache.currentGeneration;
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		return tmp->data;
	}
	void AcqisitionDataSet::updateMinMax( int generation ) {
		if ( generation == -1 )
			generation = d->cache.currentGeneration;
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		for ( int i=0; i<channels( generation ); ++i ) {
			tmp->max[ i ] = -INFINITY;
			tmp->min[ i ] = INFINITY;
			double *ch = getRawChannel( i, generation );
			for ( int j=0; j<samples( generation ); ++j ) {
				double value = ch[ j ];
				if ( value > tmp->max[ i ] ) {
					tmp->max[ i ] = value;
					addResult( QString( "channel %1/max" ).arg( i ), value );
					addResult( QString( "channel %1/max at" ).arg( i ), j );
				}
				if ( value < tmp->min[ i ] ) {
					tmp->min[ i ] = value;
					addResult( QString( "channel %1/min" ).arg( i ), value );
					addResult( QString( "channel %1/min at" ).arg( i ), j );
				}
			}
		}
	}
	void AcqisitionDataSet::resetMinMaxCache( int generation ) {
		if ( generation == -1 )
			generation = d->cache.currentGeneration;
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		for ( int i=0; i<channels( generation ); ++i ) {
			tmp->max[ i ] = -INFINITY;
			tmp->min[ i ] = INFINITY;
		}
	}

	void AcqisitionDataSet::addResult( QString name, QVariant value ) {
		//qDebug() << "AcqisitionDataSet::addResult(" << name << "," << value << ")";
		//_resulttree->debugPrint();
		NamedQVariantTreeItem* parent = d->resulttree->getNamedChild<NamedQVariantTreeItem*>( d->cache.currentPrefix );
		if ( parent == 0 )
			qWarning() << "AcqisitionDataSet::addResult(" << name << "," << value << ")" << endl << " ! ! ! Creating a new NamedQVariantTreeItem with 0 as parent.";
		while ( name.contains( "/" ) ) {
			QString firstname = name.left( name.indexOf( "/" ) );
			NamedQVariantTreeItem* tmp = parent->child<NamedQVariantTreeItem*>( firstname );
			if ( tmp == 0 )
				tmp = new NamedQVariantTreeItem( firstname, QVariant::Invalid, parent );
			parent = tmp;
			name.remove( 0, firstname.size() + 1 );
		}
		if ( parent->getQVariant( name ) )
			parent->getQVariant( name )->setValue( value );
		else
			new NamedQVariantTreeItem( name, value, parent );
	}
	NamedQVariantTreeItem* AcqisitionDataSet::result( QString name ) {
		//qDebug() << "\nAcqisitionDataSet::result(" << name << ")";
		static NamedQVariantTreeItem* invalid = new NamedQVariantTreeItem( "", QVariant::Invalid, 0 );
		if ( name.startsWith( "/data/" ) )
			name.remove( 0, 6 );
		int parents = 0;
		while ( name.startsWith( "../" ) ) {
			parents++;
			name.remove( 0, 3 );
		}
		//qDebug() << " name is becoming" << name;
		//qDebug() << " currentPrefix is" << d->cache.currentPrefix;
		NamedQVariantTreeItem* tmp = d->resulttree->getNamedChild<NamedQVariantTreeItem*>( d->cache.currentPrefix.section( "/", 0, -1-parents ) );
		//tmp->debugPrint();
		if ( tmp->getNamedChild( name ) )
			return tmp->getNamedChild<NamedQVariantTreeItem*>( name );

		//qDebug() << "AcqisitionDataSet::result(" << name << ") didn't find the desired child";
		//debugPrint();
		return invalid;
	}
	NamedQVariantTreeItem* AcqisitionDataSet::resulttree() {
		return d->resulttree;
	}

};

