/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "splinedebug.h"
#include "analyzers.h"
#include <math.h>

CFSplineDebugAnalyzer::CFSplineDebugAnalyzer( NSCore::QTreeItem* p )
	: NSEvaluation::Analyzer( p )
{
	_config->addParameter( "BackupMinimumAnalyzer", QVariant::Invalid );
	_minimumanalyzer = new InterpolatedMinimumAnalyzer( _config->getParameter( "BackupMinimumAnalyzer" ) );

	_config->addParameter( "fraction point", 0.2 );
	_config[ "fraction point" ]->setMin( 0.0 );
	_config[ "fraction point" ]->setMax( 1.0 );
}
CFSplineDebugAnalyzer::~CFSplineDebugAnalyzer() {
	delete _minimumanalyzer;
}

void CFSplineDebugAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {
	double _fraction_point = _config->getParameter( "fraction point" )->toDouble();
	for ( int i=0; i<data.channels(); ++i ) {
		NSCore::ChannelDataSet* ch = data.channel( i );
		QVariant* _min_parameter = ch->result( "interpolated minimum" );
		QVariant* _min_at_parameter = ch->result( "interpolated minimum at" );
		if ( _min_parameter == 0 || _min_at_parameter == 0 ) {
			qDebug() << " running own Analyzer";
			_minimumanalyzer->analyze( data );
			_min_parameter = ch->result( "interpolated minimum" );
			_min_at_parameter = ch->result( "interpolated minimum at" );
		}
		double _minimum = ( ( _min_parameter )?_min_parameter->toDouble():0 );
		double _minimum_at = ( ( _min_at_parameter )?_min_at_parameter->toDouble():0 );

		if ( _minimum_at > 10 && _minimum_at < ch->samples() ) {
			double zeroline=0;
			for ( int i=1; i<_minimum_at-10; ++i )
				zeroline += ch->sample( i );
			zeroline /= _minimum_at-11;

			double _fraction = zeroline - ( zeroline - _minimum ) * _fraction_point;

			// Find the first sample to exceed _fraction
			bool _updown = false;
			if ( ch->sample( 2 ) > _fraction )
				_updown = true;

			int _fraction_at_int = 0;
			for ( int i=/*int( _minimum_at )-*/10; i<int( _minimum_at ) && _fraction_at_int==0; ++i ) {
				if ( _updown && ch->sample( i ) < _fraction )
					_fraction_at_int = i;
				if ( !_updown && ch->sample( i ) > _fraction )
					_fraction_at_int = i;
			}

			//setMiddle( _fraction_at_int );
			setData( ch->sampledata(), ch->samples() );
			setDerivates( 1e40, 1e40 );

			double x0,x1,x2, y0,y1,y2;
			x0 = _fraction_at_int-1;
			y0 = splint( _fraction_at_int-1 );
			x1 = _fraction_at_int;
			y1 = splint( _fraction_at_int );
			x2 = _fraction_at_int+1;
			y2 = splint( _fraction_at_int+1 );

			while ( fabs( y1 - _fraction ) > 1E-4 && x0 != x2 ) {
				if ( y1 < _fraction && y0 > _fraction ) {
					x2 = x1;
					y2 = y1;
				} else {
					x0 = x1;
					y0 = y1;
				}
				double newx = x0 + ( x2 - x0 )/2;
				x1 = newx;
				y1 = splint( newx );
			}

			//qDebug() << "trying to add a result (int. fraction point at =" << x1 << ")";
			//qDebug() << " to channel" << ch;
			ch->addResult( "interpolated fraction point at", x1 );
			ch->addResult( "interpolated fraction point", y1 );
			ch->addResult( "zeroline", zeroline );
			double error = fabs( ( y1 - _fraction ) /* / _fraction */ );
			if ( error > 2 ) {
				qDebug() << "\tCF-error =" << error;
				qDebug() << "\t\t_fraction =" << _fraction << " points[1].y =" << y1;
				ch->debugPrint();
			}
			ch->addResult( "error", error );
		}
	}
}
#define AVERAGE 0
double CFSplineDebugAnalyzer::y( int n ) {
#ifndef AVERAGE
	return _y[ n ];
#else
	double ret = 0;
	int m = 0;
	if ( n>0 ) {
		ret += _y[ n-1 ];
		m++;
	}
	ret += 2*_y[ n ];
	m+=2;
	if ( n<_n ) {
		ret += _y[ n+1 ];
		m++;
	}
	return ret / m;
#endif
}

